VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "brMatObj"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
' CAPE-OPEN Unit Bridge (c) 2001 Joerg Koeller

' CAUTION: This software comes without any warranty whatsoever.
' You use this software at your very own risk. The author is
' is not responsible for any damage cause directly or indirectly
' by the use of this software.

' This software may be used for educational purposes only. Any
' commercial use neither as source code nor as binary even in
' parts or modified form is strictly
' prohibited and requires a written permission by the author.
' You can contact the author via e-mail: mandrox@web.de


Option Explicit

Implements CAPEOPEN.ICapeIdentification
Implements CAPEOPEN.ICapeThermoMaterialObject

Private cbObjRef As cOrbObjRef
Private cbObject As c_MatObj


' I CORBA IDL v0.9.2 ICapeThermoMaterialObject does not support ICapeIdentification
' Therefore, the bridge has to manage it's name and description by itself

Private myComponentName As String
Private myComponentDescription As String

Friend Sub initCB(externalRef As String)

    If (myOrb Is Nothing) Then
        Set myOrb = VBOrb.init
    End If
    
    Set cbObjRef = retrieveObject(externalRef)
    
    Set cbObject = m_MatObj.narrow(cbObjRef)
           
End Sub

Friend Sub initCBobj(cbObj As cOrbObject)
    Set cbObject = m_MatObj.narrow(cbObj)
End Sub

Private Sub Class_Initialize()
    myComponentName = "COM-CORBA bridge Material Object"
    myComponentDescription = "This component provides a COM v0.9.3 wrapper for an CORBA v0.9.2 Material Object"
End Sub

Private Property Let ICapeIdentification_ComponentDescription(ByVal RHS As String)
    myComponentDescription = RHS
End Property

Private Property Get ICapeIdentification_ComponentDescription() As String
    ICapeIdentification_ComponentDescription = myComponentDescription
End Property

Private Property Let ICapeIdentification_ComponentName(ByVal RHS As String)
    myComponentName = RHS
End Property

Private Property Get ICapeIdentification_ComponentName() As String
    ICapeIdentification_ComponentName = myComponentName
End Property

Private Function ICapeThermoMaterialObject_AvailableProps() As Variant
    Dim cbProps As c_StringSeq
    Set cbProps = cbObject.GetAvailableProps
    Dim rVal As Variant
    rVal = stringSeq2Variant(cbProps)
    ICapeThermoMaterialObject_AvailableProps = rVal
End Function

Private Sub ICapeThermoMaterialObject_CalcEquilibrium(ByVal flashType As String, ByVal props As Variant)
   Dim cbProps As c_StringSeq
   Set cbProps = variant2StringSeq(props)
   cbObject.CalcEquilibrium flashType, cbProps
End Sub

Private Sub ICapeThermoMaterialObject_CalcProp(ByVal props As Variant, ByVal phases As Variant, ByVal calcType As String)
   Dim cbProps As c_StringSeq
   Set cbProps = variant2StringSeq(props)
   Dim cbPhases As c_StringSeq
   Set cbPhases = variant2StringSeq(phases)
   cbObject.CalcProp cbProps, cbPhases, calcType
End Sub

Private Property Get ICapeThermoMaterialObject_ComponentIds() As Variant
'Public Property Get ComponentIds() As Variant
    Dim cbComponents As c_StringSeq
    Set cbComponents = cbObject.GetComponentIds
    Dim rVal As Variant
    rVal = stringSeq2Variant(cbComponents)
    ICapeThermoMaterialObject_ComponentIds = rVal
    'ComponentIds = rVal
End Property

Private Function ICapeThermoMaterialObject_CreateMaterialObject() As Object
   ' CORBA v0.9.2 does not support thre CreateMaterialObject Method.
   ' Therefore the duplicate method is called instead
   
    Dim newCBMo As c_MatObj
    Set newCBMo = cbObject.Duplicate
    
    Dim newCCMo As brMatObj
    Set newCCMo = New brMatObj 'CreateObject("unitBridge.brMatObj")
    newCCMo.initCBobj newCBMo
    
    Set ICapeThermoMaterialObject_Duplicate = newCCMo
End Function

Private Function ICapeThermoMaterialObject_Duplicate() As Object
    Dim newCBMo As c_MatObj
    Set newCBMo = cbObject.Duplicate
    
    Dim newCCMo As brMatObj
    Set newCCMo = New brMatObj 'CreateObject("unitBridge.brMatObj")
    newCCMo.initCBobj newCBMo
    
    Set ICapeThermoMaterialObject_Duplicate = newCCMo
End Function

Private Function ICapeThermoMaterialObject_GetComponentConstant(ByVal props As Variant, ByVal compIds As Variant) As Variant
'Public Function GetComponentConstant(ByVal props As Variant, ByVal compIds As Variant) As Variant
    Dim cbProps As c_StringSeq
    Set cbProps = variant2StringSeq(props)
    Dim cbCompIds As c_StringSeq
    Set cbCompIds = variant2StringSeq(compIds)
    Dim cbrVal As c_anySeq
    Set cbrVal = cbObject.GetComponentConstant(cbProps, cbCompIds)
    Dim rVal As Variant
    rVal = anySeq2Variant(cbrVal)
    ICapeThermoMaterialObject_GetComponentConstant = rVal
    'GetComponentConstant = rVal
End Function

Private Function ICapeThermoMaterialObject_GetIndependentVar(ByVal indVars As Variant) As Variant
'Public Function GetIndependentVar(ByVal indVars As Variant) As Variant
    Dim cbVars As c_StringSeq
    Set cbVars = variant2StringSeq(indVars)
    Dim cbVals As c_DoubleSeq
    Set cbVals = cbObject.GetIndependentVar(cbVars)
    Dim rVal As Variant
    rVal = doubleSeq2Variant(cbVals)
    ICapeThermoMaterialObject_GetIndependentVar = rVal
    'GetIndependentVar = rVal
End Function

Private Function ICapeThermoMaterialObject_GetNumComponents() As Long
    ICapeThermoMaterialObject_GetNumComponents = cbObject.GetNumComponents
End Function

Private Function ICapeThermoMaterialObject_GetProp(ByVal property As String, ByVal phase As String, ByVal compIds As Variant, ByVal calcType As String, ByVal basis As String) As Variant
    Dim cbCompIds As c_StringSeq
    Set cbCompIds = variant2StringSeq(compIds)
    Dim result As c_DoubleSeq
    Set result = cbObject.GetProp(property, phase, cbCompIds, calcType, basis)
    Dim rVal As Variant
    rVal = doubleSeq2Variant(result)
    ICapeThermoMaterialObject_GetProp = rVal
End Function

Private Function ICapeThermoMaterialObject_GetPropList() As Variant
' CORBA IDL v0.9.2 does not support the get PropListMethof. Calling GetAvailableProps
' To be changed in next version of the bridge
    Dim stringSeq As c_StringSeq
    Set stringSeq = cbObject.GetAvailableProps()
    Dim rVal As Variant
    rVal = stringSeq2Variant(stringSeq)
    ICapeThermoMaterialObject_GetPropList = rVal
End Function

Private Function ICapeThermoMaterialObject_GetUniversalConstant(ByVal props As Variant) As Variant
   Dim cbProps As c_StringSeq
   Set cbProps = variant2StringSeq(props)
   Dim cbConsts As c_anySeq
   Set cbConsts = cbObject.GetUniversalConstant(cbProps)
   Dim rVal As Variant
   rVal = anySeq2Variant(cbConsts)
   ICapeThermoMaterialObject_GetUniversalConstant = rVal
End Function

Private Property Get ICapeThermoMaterialObject_PhaseIds() As Variant
'Public Property Get PhaseIds() As Variant
  Dim cbPhases As c_StringSeq
  Set cbPhases = cbObject.GetPhaseIds()
  Dim rVal As Variant
  rVal = stringSeq2Variant(cbPhases)
  'PhaseIds = rVal
  ICapeThermoMaterialObject_PhaseIds = rVal
End Property

Private Function ICapeThermoMaterialObject_PropCheck(ByVal props As Variant) As Variant
  Dim cbProps As c_StringSeq
  Set cbProps = variant2StringSeq(props)
  Dim cbBools As c_BooleanSeq
  Set cbBools = cbObject.PropCheck(cbProps)
  Dim rVal As Variant
  rVal = booleanSeq2Variant(cbBools)
  ICapeThermoMaterialObject_PropCheck = rVal
End Function

Private Sub ICapeThermoMaterialObject_RemoveResults(ByVal props As Variant)
  
' Implementation should look like this nut CIDL v0.9.2 does not support this
' Method. Therefore raising an NoImplement error
'  Dim cbProps As c_StringSeq
'  Set cbProps = variant2StringSeq(props)
'  cbObject.RemoveResults cbProps

    Err.Raise vbObjectError

End Sub

Private Sub ICapeThermoMaterialObject_SetIndependentVar(ByVal indVars As Variant, ByVal values As Variant)
'Public Sub SetIndependentVar(ByVal indVars As Variant, ByVal values As Variant)
    Dim cbVars As c_StringSeq
    Set cbVars = variant2StringSeq(indVars)
    Dim cbVals As c_DoubleSeq
    Set cbVals = variant2DoubleSeq(values)
    cbObject.SetIndependentVar cbVars, cbVals
End Sub

Private Sub ICapeThermoMaterialObject_SetProp(ByVal property As String, ByVal phase As String, ByVal compIds As Variant, ByVal calcType As String, ByVal basis As String, ByVal values As Variant)
    Dim cbCompIds As c_StringSeq
    Set cbCompIds = variant2StringSeq(compIds)
    Dim cbValues As c_DoubleSeq
    Set cbValues = variant2DoubleSeq(values)
    cbObject.SetProp property, phase, cbCompIds, calcType, basis, cbValues
End Sub

Private Function ICapeThermoMaterialObject_ValidityCheck(ByVal props As Variant) As Variant
' The return type (ICapeThermoReliability) of this function is not specified so far in the
' CAPE-OPEN specs. Therefore this implmementation was intentionally left blank :))
    
    Err.Raise vbObjectError
End Function
